Replaces hardcoded thresholds AND clamp bounds across 12 RL controllers
with observed-signal-driven ISV-slot bounds. Eliminates the architectural
failure mode that surfaced in walk-forward fold 0 (alpha-rl-m9cx5) and
fold 1 (alpha-rl-jgdh6): under Phase 4.5 advantage normalization, eight
controllers (PPO clip, target_tau, rollout_steps, entropy_coef, per_α,
gamma, reward_scale, q_distill_lambda) saturated at extrema within ~50
steps and stayed pinned for the rest of training — same pattern across
two different data slices, confirming the bug is structural rather than
data-size-dependent.
Mechanism: each controller's noise floor was hardcoded as a small
fraction of its target (`_NOISE_FLOOR_FRAC = 0.01f`), calibrated against
a pre-Phase-4.5 signal regime. Phase 4.5 normalization reduces operating
KL by ~50× — observed signal stays below the 1%-of-target floor, the
Schulman widen path fires continuously (asymmetric in the wrong
direction), and ε hits MAX 0.50 within ~50 training steps. ksll2's full-
data run (n_folds=1) happened to escape via a single above-band KL
observation that triggered tighten; both walk-forward folds (3 and 6
files) did not.
Per `feedback_adaptive_not_tuned`, `feedback_isv_for_adaptive_bounds`,
`pearl_controller_anchors_isv_driven`: every threshold now derives from
observed signal statistics (Welford online variance) rather than
constants calibrated against a prior signal regime. User explicitly
expanded scope mid-implementation: "if all clamps ISV bound should be
added to this spec and tasks" — applying the principle consistently
means hardcoded MIN/MAX clamp bounds count too, not just the saturating
noise floors. 12 controllers, single atomic commit per
`feedback_no_partial_refactor`.
Spec: docs/superpowers/specs/2026-05-30-adaptive-controller-floor-design.md
Plan: docs/superpowers/plans/2026-05-30-adaptive-controller-floor-plan.md
# New kernel
- rl_signal_variance_update.cu (80 LOC) — Welford online variance.
Single-thread single-block. Sentinel-zero skip per pearl. Per-controller
Welford triple (count, mean, M²) drives every adaptive noise floor.
# Per-controller refactors (12 .cu files)
- ppo_clip + target_tau + rollout_steps + entropy_coef + per_α —
adaptive noise floor = max(target × 0.5, sqrt(observed_var) × 2);
asymmetric Schulman (tighten on single observation, widen requires 3
consecutive below-band).
- gamma (Special G) — hardcoded GAMMA_MIN = 0.995 → adaptive via
Welford MEAN of trade duration. Per spec Q6 resolution: the Welford
mean's natural N-smoothing lag breaks the gamma↔trade_duration
feedback loop without explicit step-period gating. 100-observation
warmup falls back to EMA before Welford has enough samples.
- reward_scale (Special R) — asymmetric DECREASE rate cap (5% per
step) on both bootstrap-replace and Wiener-blend paths; bootstrap-
fraction floor (10% of bootstrap) until 100 trades close.
- q_distill_lambda (Special Q) — hardcoded MIN_LAMBDA = 0.05 → adaptive
via Welford on q_distill_kl_ema: max(0.001, std × 0.05).
- v_blend_alpha (Phase 4.4) — 5 hardcoded constants → 5 ISV slots
(DEAD_SIGNAL_FLOOR, TARGET_TRACK_RATIO, SCHULMAN_STEP, EMA_ALPHA,
BOOTSTRAP_ALPHA) + adaptive dead-signal floor from V_scalar magnitude
variance.
- ppo_ratio_clamp — adaptive MIN/MAX from observed log-ratio variance.
Architectural 2.0 absolute floor preserved (don't degenerate to
vanilla policy gradient).
- reward_clamp — V_BOUND_FLOOR, V_BOUND_EWMA_ALPHA, MIN_WIN, MIN_RATIO,
MAX_RATIO, MIN/MAX_MARGIN, MARGIN_TOLERANCE/ADJUST_RATE,
CLIP_RATE_EMA_ALPHA → 10 ISV slots.
- gate_threshold — hardcoded alpha = 0.01 → ISV slot 638.
- Clamp-bound expansion: EPS_MIN/MAX, TAU_MIN, ROLLOUT_MIN/MAX,
COEF_MIN/MAX, PER_ALPHA_MIN/MAX, GAMMA_MAX, MAX_LAMBDA, KL_TOLERANCE,
LAMBDA_RAMP_RATE, LAMBDA_DECAY_RATE → all ISV.
- WIENER_ALPHA_FLOOR shared across 9 controllers → single ISV slot 659.
# Trainer integration (integrated.rs)
- rl_signal_variance_update kernel loaded + helper method
`launch_rl_signal_variance_update`.
- Per-step Welford launches for 9 controller inputs, placed between
EMA producers and rl_fused_controllers in the per-step pipeline.
- Input-slot lookup array for rl_fused_controllers updated: rollout_steps
now consumes RL_ADV_VAR_PRE_NORM_INDEX (emitted by
rl_advantage_normalize before in-place normalize) instead of the
post-norm advantage_var_ratio (definitionally ~0 under Phase 4.5).
- ~30 new ISV bootstrap entries in with_controllers_bootstrapped.
# ISV slot allocation (isv_slots.rs)
- 72 new slots, RL_SLOTS_END 588 → 660. +288 bytes mapped-pinned.
- 9 Welford variance triples + 5 asymmetric Schulman counters +
3 new input signals (var_pre_norm, gamma_min_adaptive, reward_mag) +
v_blend (5 + 3 Welford) + ppo_ratio_clamp (1 + 3 Welford) +
reward_clamp (5) + gate_threshold (1) + q_distill (1) + 20 clamp bounds +
shared wiener floor.
# Bootstrap-clamp consistency fix (caught by Task 16 testing)
The original draft bootstrapped RL_EPS_BOOTSTRAP at 0.01 (= KL target),
but EPS_MIN was 0.05 — bootstrap value below clamp range. First post-
bootstrap step always snapped ε to MIN regardless of signal direction
("snap to MIN" behavior the unit tests surfaced). Fixed by bootstrapping
ε at MIN (0.05) so asymmetric Schulman operates from a valid state.
# Validation
- cargo build --release: clean
- cargo build --tests: clean
- 3 GPU Welford kernel tests (G1 constant→0var, G2 sequence→known var,
G3 sentinel skip): all pass
- 5 GPU adaptive-floor invariant tests (G3 PPO clip holds at bootstrap
when signal below floor, G4 tightens on single above-band, G5 widens
only after 3 consecutive below-band, G6 post-warmup uses Welford mean,
G6 pre-warmup uses EMA): all pass
- integrated_trainer_smoke (full GPU pipeline): passes
- 1k local smoke at b=128: 1000/1000 steps, completed_clean, no NaN,
controllers genuinely adapting (γ 0.90→0.974, per_α 0.40→0.52,
q_distill_λ 0.05→0.21, ε held at 0.05 = MIN per Phase 4.5 small-KL
regime — was previously stuck at MAX 0.50 throughout fold 0/1)
- compute-sanitizer memcheck b=128 5 steps: 0 errors
Cluster validation (G7-G9) submitted as follow-up runs.
🤖 Generated with [Claude Code](https://claude.com/claude-code)
Co-Authored-By: Claude Opus 4.7 <noreply@anthropic.com>
88 lines
3.6 KiB
Plaintext
88 lines
3.6 KiB
Plaintext
// rl_gate_threshold_controller.cu — adaptive gate threshold controller.
|
||
//
|
||
// Watches trade frequency (dones EMA) and adjusts confidence + FRD
|
||
// gate thresholds to maintain a target trade rate. When trades dry up
|
||
// (dones_ema < target), thresholds DECREASE (more permissive). When
|
||
// trades are abundant (dones_ema > target), thresholds INCREASE
|
||
// (more selective). Schulman bounded discrete step pattern.
|
||
//
|
||
// Runs once per step, single thread. Reads/writes ISV.
|
||
// Per `feedback_no_atomicadd`, `feedback_cpu_is_read_only`.
|
||
|
||
#include <stdint.h>
|
||
|
||
#define RL_CONF_GATE_THRESHOLD_INDEX 512
|
||
#define RL_FRD_GATE_THR_LONG_INDEX 516
|
||
#define RL_FRD_GATE_THR_SHORT_INDEX 517
|
||
#define RL_GATE_WARMUP_STEPS_INDEX 524
|
||
#define RL_GATE_DONES_EMA_INDEX 525
|
||
#define RL_GATE_DONES_TARGET_INDEX 526
|
||
#define RL_GATE_CONF_MIN_INDEX 527
|
||
#define RL_GATE_CONF_MAX_INDEX 528
|
||
#define RL_GATE_FRD_MIN_INDEX 529
|
||
#define RL_GATE_FRD_MAX_INDEX 530
|
||
#define RL_GATE_ADJUST_RATE_INDEX 531
|
||
#define RL_STEP_COUNTER_ISV_INDEX 548
|
||
// Adaptive controller floors (spec 2026-05-30-adaptive-controller-floor-design):
|
||
// EMA smoothing α was hardcoded 0.01f; ISV slot 638 makes it tunable
|
||
// at runtime so the dones-rate EMA half-life can adapt to data volume
|
||
// (e.g. small-batch smoke runs may want a faster α to surface signal
|
||
// before warmup completes).
|
||
#define RL_GATE_EMA_ALPHA_INDEX 638
|
||
|
||
extern "C" __global__ void rl_gate_threshold_controller(
|
||
float* __restrict__ isv,
|
||
const float* __restrict__ dones, // [B]
|
||
int b_size
|
||
) {
|
||
if (threadIdx.x != 0 || blockIdx.x != 0) return;
|
||
|
||
const int current_step = (int)isv[RL_STEP_COUNTER_ISV_INDEX];
|
||
const int warmup = (int)isv[RL_GATE_WARMUP_STEPS_INDEX];
|
||
if (current_step < warmup) return;
|
||
|
||
// Update dones EMA (count of done events this step / b_size).
|
||
float done_count = 0.0f;
|
||
for (int b = 0; b < b_size; ++b) {
|
||
if (dones[b] > 0.5f) done_count += 1.0f;
|
||
}
|
||
const float done_rate = done_count / (float)b_size;
|
||
const float alpha = isv[RL_GATE_EMA_ALPHA_INDEX];
|
||
const float prev_ema = isv[RL_GATE_DONES_EMA_INDEX];
|
||
const float dones_ema = (prev_ema == 0.0f && done_rate > 0.0f)
|
||
? done_rate
|
||
: (1.0f - alpha) * prev_ema + alpha * done_rate;
|
||
isv[RL_GATE_DONES_EMA_INDEX] = dones_ema;
|
||
|
||
const float target = isv[RL_GATE_DONES_TARGET_INDEX];
|
||
const float adjust = isv[RL_GATE_ADJUST_RATE_INDEX];
|
||
|
||
// Schulman bounded step: if dones_ema < target, DECREASE thresholds
|
||
// (more permissive); if > target, INCREASE (more selective).
|
||
float conf_thr = isv[RL_CONF_GATE_THRESHOLD_INDEX];
|
||
float frd_long = isv[RL_FRD_GATE_THR_LONG_INDEX];
|
||
float frd_short = isv[RL_FRD_GATE_THR_SHORT_INDEX];
|
||
|
||
if (dones_ema < target * 0.2f) {
|
||
// Trades well below target — relax gates.
|
||
conf_thr *= adjust;
|
||
frd_long *= adjust;
|
||
frd_short *= adjust;
|
||
} else if (dones_ema > target * 5.0f) {
|
||
// Trades well above target — tighten gates.
|
||
conf_thr /= adjust;
|
||
frd_long /= adjust;
|
||
frd_short /= adjust;
|
||
}
|
||
|
||
// Bilateral clamp.
|
||
const float conf_min = isv[RL_GATE_CONF_MIN_INDEX];
|
||
const float conf_max = isv[RL_GATE_CONF_MAX_INDEX];
|
||
const float frd_min = isv[RL_GATE_FRD_MIN_INDEX];
|
||
const float frd_max = isv[RL_GATE_FRD_MAX_INDEX];
|
||
|
||
isv[RL_CONF_GATE_THRESHOLD_INDEX] = fmaxf(conf_min, fminf(conf_thr, conf_max));
|
||
isv[RL_FRD_GATE_THR_LONG_INDEX] = fmaxf(frd_min, fminf(frd_long, frd_max));
|
||
isv[RL_FRD_GATE_THR_SHORT_INDEX] = fmaxf(frd_min, fminf(frd_short, frd_max));
|
||
}
|